/*
 * Copyright (C) 2010 Company 100, Inc. All rights reserved.
 * Copyright (C) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"


#include <wtf/Platform.h> // PLATFORM(...)

#if OS(WINDOWS) || OS(UNIX)
#include <stdio.h>
#define OSReport printf
#endif

#if OS(WINDOWS)
#include <winsock2.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#endif 

#if OS(UNIX)
#include <netinet/in.h>
#include <netdb.h>
#endif

#include "WiiUEventManager.h"
#include "wk_event.h"
#include "KURL.h"
#include "Logging.h"
#include "NotImplemented.h"
#include "SocketStreamHandleClient.h"
#include "WebSocketStreamHandle.h"
#include "WebSocketStreamHandlePrivate.h"
#include <wtf/Vector.h>
#include <wtf/text/CString.h>

#if PLATFORM(WIIU)
#include "DomainAccessControlWiiU.h"
    #if defined(ENABLE_HIO_INSPECTOR)
        #include <HTTPParsers.h>
    #endif
    
    #include <PlatformOS.h>
    #include <PlatformDefines.h>
	#include <PlatformError.h>
    using namespace LibPlatform;
#endif

namespace WebCore {

TcpSocketWiiU::TcpSocketWiiU(WebSocketStreamHandlePrivate * handle )
    : m_receiveTimer(this, &TcpSocketWiiU::notificationReceiveTimerFired)
    , m_handle(handle)
{
   m_receiveTimer.startRepeating(0.2);
}
TcpSocketWiiU::TcpSocketWiiU()
    : m_receiveTimer(this, &TcpSocketWiiU::notificationReceiveTimerFired)
    , m_handle(NULL)
{
}

TcpSocketWiiU::~TcpSocketWiiU()
{
    m_receiveTimer.stop();
}
void TcpSocketWiiU::notificationReceiveTimerFired(Timer<TcpSocketWiiU>*)
{
    if (m_handle)
    {
        if (m_handle->isConnected()) {
            int ret = m_handle->socketReadyRead();
            if (ret <= 0) {
				#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
				if (m_handle->m_hioHandle != DEBUGIO_HANDLE_INVALID) {
					if (ret < 0) {
						m_receiveTimer.stop();
						m_handle->socketClosed();
					}
					return;
				}
				#endif
				#if OS(WINDOWS)
				int error = WSAGetLastError(); 
				if(ret < 0 && (error == WSAEWOULDBLOCK || error == 0))
				#else 
				int error = errno; 
				if(error == EWOULDBLOCK)
				#endif
				{
					return; 
				}
                m_receiveTimer.stop();
				m_handle->socketClosed(); 
            }
        } else {
            if (m_handle->isClosed()) {
                 m_receiveTimer.stop();
                 m_handle->socketClosed();
            } else
            m_handle->socketConnected();
        }
    }
}

#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
bool WebSocketStreamHandlePrivate::m_hioConnected = false;
WTF::Mutex *WebSocketStreamHandlePrivate::m_hioConnectedMutexP;
static char hioBuffer[1024] = { 0 }; // Temporary read buffer for the async write
static bool hioReading = false;
static WTF::Vector<char> *hioReadBufferP; // "non-blocking" HIO read buffer
static WTF::Mutex *hioMutexP;

void ResetWebSocketStreamHandleWiiU(void)
{
	WebSocketStreamHandlePrivate::m_hioConnected = false;
	WebSocketStreamHandlePrivate::m_hioConnectedMutexP = new WTF::Mutex;
	memset(hioBuffer, 0, sizeof(hioBuffer));
	hioReading = false;
	hioReadBufferP = new WTF::Vector<char>; // "non-blocking" HIO read buffer
	hioMutexP = new WTF::Mutex;
}

static int WebSocket_HIOReadAsync(DebugIO::Handle handle);
static void hioReadCallback(DebugIO::Status status, void *context)
{
    DebugIO::Handle hioHandle = (DebugIO::Handle)context;
    (void)hioHandle; // Only one HIO handle for the inspector anyway

    hioMutexP->lock();
    if (!hioReading) {
        hioMutexP->unlock();
        return;
    }

    if (status > 0) hioReadBufferP->append(hioBuffer, (size_t)status);
    hioReading = false;
    hioMutexP->unlock();


    // Fixme: It seems the Host I/O recv buffer is exactly 500 bytes, so this could in theory facilitate larger transfers
    //        Testing so far has shown that it isn't really necessary though
    //        - Enabling this because of bug #18376
    if (status == 500) while (WebSocket_HIOReadAsync(hioHandle) == DEBUGIO_STATUS_NO_CLIENT_TXN_BUF_AVAILABLE);
}

static int WebSocket_HIORecv(char *buffer, int len)
{
    // If nothing is connected by HIO, assume the connection fails so HIO sockets can be closed properly
    WebSocketStreamHandlePrivate::m_hioConnectedMutexP->lock();
    bool hioConnected = WebSocketStreamHandlePrivate::m_hioConnected;
    WebSocketStreamHandlePrivate::m_hioConnectedMutexP->unlock();
    if (!hioConnected) return -1;

    if (len == 0) return 0;

    hioMutexP->lock();
    int actualRead = len;
    if (hioReadBufferP->size() < (uint32_t)actualRead) actualRead = (int)hioReadBufferP->size();
    if (actualRead > 0) {
        memcpy(buffer, &(*hioReadBufferP)[0], actualRead);

        // Fixme: This blob parses a full HTTP GET request for the inspectorPageIndex.html file without going into the next request
        //        Might be useful later?
        //static String failureReason;
        //static String tempMethod;
        //static String tempUrl;
        //static HTTPVersion tempHttpVersion;
        //static AtomicString tempNameStr;
        //static String tempValueStr;
        //size_t httpEnd = parseHTTPRequestLine(buffer, actualRead, failureReason, tempMethod, tempUrl, tempHttpVersion);
        //if (httpEnd == 0)
        //    actualRead = 0;
        //else {
        //    size_t httpHeaderEnd;
        //    do {
        //        httpHeaderEnd = parseHTTPHeader(buffer + httpEnd, actualRead - httpEnd, failureReason, tempNameStr, tempValueStr);
        //        if (httpHeaderEnd == 0) actualRead = 0;
        //        else {
        //            httpEnd += httpHeaderEnd + 1;
        //            if (tempNameStr.isEmpty()) {
        //                actualRead = (int)httpEnd - 1;
        //                break;
        //            }
        //        }
        //    } while (httpHeaderEnd != 0);
        //}

        if (actualRead > 0) hioReadBufferP->remove(0, actualRead);
    }
    hioMutexP->unlock();

    return actualRead;
}

static int WebSocket_HIORecvMultiple(WebSocketStreamHandle *stream, char *buffer, int len) {
    // Fixme: Inspector over HIO really only has one handle, so sends and recvs on different sockets 
    //         all go to the same place and we can deal with all the HTTP GET requests as soon as we get them
    //        Of course to do this, we have to make sure the shutdownAfterSendOrNow call in WebInspectorServer.cpp
    //         WebInspectorServer::didReceiveUnrecognizedHTTPRequest() is avoided so the socket isn't closed prematurely,
    //         and then we have to close it manually to avoid a leak
    //int totalRecv = 0;
    int curRecv = 0;
    //bool didSendOnce = false;
    //do {
           curRecv = WebSocket_HIORecv(buffer, len);
           //if (curRecv > 0) {
              // stream->client()->didReceiveSocketStreamData(stream, buffer, curRecv);
              // didSendOnce = true;
           //}
           //else if (didSendOnce) stream->client()->didReceiveSocketStreamData(stream, NULL, -1);
           //totalRecv += curRecv;
    //} while (curRecv > 0);
    //return 0;
    return curRecv;
}

static int WebSocket_HIOReadAsync(DebugIO::Handle handle)
{
    hioMutexP->lock();
    int ret = 0;
    if (!hioReading) {
        hioReading = true;
        DebugIO::Status hioResult = DebugIO::ReadAsync(handle, sizeof(hioBuffer), (void*)hioBuffer, hioReadCallback, (void*)handle);
        if (hioResult < 0 && hioResult != DEBUGIO_STATUS_NO_CLIENT_TXN_BUF_AVAILABLE) ret = -1;
    }
    hioMutexP->unlock();
    return ret;
}
#else
void ResetWebSocketStreamHandleWiiU(void)
{
}
#endif

WebSocketStreamHandlePrivate::WebSocketStreamHandlePrivate(WebSocketStreamHandle* streamHandle, int fd)
    : TcpSocketWiiU(this)
    , m_streamHandle(streamHandle)
    , m_socketFd(fd)
    #if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
    , m_hioHandle(DEBUGIO_HANDLE_INVALID)
    #endif
{
}

#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
WebSocketStreamHandlePrivate::WebSocketStreamHandlePrivate(WebSocketStreamHandle* streamHandle, int fd, bool isHio)
    : TcpSocketWiiU(this)
    , m_streamHandle(streamHandle)
    , m_socketFd(fd)
    , m_hioHandle(DEBUGIO_HANDLE_INVALID)
{
    if (isHio) m_hioHandle = (DebugIO::Handle)fd;
}
#endif

WebSocketStreamHandlePrivate::~WebSocketStreamHandlePrivate()
{
}

void WebSocketStreamHandlePrivate::socketConnected()
{
    if (m_streamHandle && m_streamHandle->client() && m_streamHandle->m_state == SocketStreamHandleBase::Connecting) {
        m_streamHandle->m_state = SocketStreamHandleBase::Open;
        m_streamHandle->client()->didOpenSocketStream(m_streamHandle);
    }
}

bool WebSocketStreamHandlePrivate::isConnected()
{
    if ((m_streamHandle->m_state == SocketStreamHandleBase::Open)
       || (m_streamHandle->m_state == SocketStreamHandleBase::Closing)) {
        return true;
    }
    return false;
}

bool WebSocketStreamHandlePrivate::isClosed()
{
    if (m_streamHandle->m_state == SocketStreamHandleBase::Closed) {
        return true;
    }
    return false;
}

int WebSocketStreamHandlePrivate::socketReadyRead()
{
    int readSize = 0;
    char buffer[8192]; //CRANK-ME correct size ? 
    if (m_streamHandle && m_streamHandle->client()) {
        #if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
        if (m_hioHandle != DEBUGIO_HANDLE_INVALID) {
            if (WebSocket_HIOReadAsync(m_hioHandle) < 0) return -1;
            readSize = WebSocket_HIORecvMultiple(m_streamHandle, buffer, sizeof(buffer));
            if (readSize > 0) {
                m_streamHandle->client()->didReceiveSocketStreamData(m_streamHandle, (const char*)buffer, readSize);
            }
            return readSize;
        }
        #endif

        #if OS(WINDOWS)
        u_long blocking=1;
        ioctlsocket(m_socketFd,FIONBIO,&blocking);
        readSize = recv(m_socketFd, buffer, sizeof(buffer), 0);
        blocking=0;
        ioctlsocket(m_socketFd,FIONBIO,&blocking);
        #else
        readSize = recv(m_socketFd, buffer, sizeof(buffer), MSG_DONTWAIT);
        #endif

        if (readSize == 0) {
            return 0;
        } else if (readSize  < 0) {
            return -1;
        }
    //    OS::Report("Received %d bytes\n", readSize);
        m_streamHandle->client()->didReceiveSocketStreamData(m_streamHandle, (const char*)buffer, readSize);
   }
   return readSize;
}


int WebSocketStreamHandlePrivate::socketSend(const char* data, int len)
{
    int sendSize=0;
    int ret = 0; 

    #if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
    if (m_hioHandle != DEBUGIO_HANDLE_INVALID) {
        if (m_streamHandle && m_streamHandle->client()) {
            // Assume the write failed if there's nothing connected by HIO anymore
            m_hioConnectedMutexP->lock();
            bool hioConnected = m_hioConnected;
            m_hioConnectedMutexP->unlock();
            if (!hioConnected) return -1;

            while (sendSize != len) {
                ret = (int)DebugIO::Write(m_hioHandle, (unsigned)(len - sendSize), (const void*)(data + sendSize));
                if (ret == DEBUGIO_STATUS_NO_CLIENT_TXN_BUF_AVAILABLE) continue;
                else if (ret < 0) return -1;
                sendSize += ret;
            }
            // Fixme: So apparently you need some kind of OS::Report with a decently sized string or inspector ends up just hanging
            //        My guess is that OS::Report runs over the same mechanism as HIO, and without this OS::Report call, the HIOWrite
            //         gets buffered and is never actually sent, resulting in an inspector hang
            //OS::Report("HIO inspector write\n");
            m_streamHandle->sendPendingData();
        }

        return sendSize;
    }
    #endif

    if (m_socketFd == -1) {
	    OS::Report("Socket Creation Fail error: %d\n", errno);
	    return -1;
    }

	if (m_streamHandle && m_streamHandle->client()) {
	    while (sendSize != len) {
	        ret = send(m_socketFd, &data[sendSize], len-sendSize, 0);
        	if (ret < 0) {
				#if OS(WINDOWS)
				int error = WSAGetLastError();
				#else
            	int error = errno;
				#endif
	            OS::Report("Socket Send Fail error: %d\n", error);
	            return -1;
	        }
	        sendSize += ret;
	    }
	    m_streamHandle->sendPendingData();
	} else {
	    //OS::Report("ERROR NO socket stream or handle\n");
	}

    return sendSize;
}

void WebSocketStreamHandlePrivate::socketClose()
{
    #if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
    if (m_hioHandle != DEBUGIO_HANDLE_INVALID) return;
    #endif
    #if OS(WINDOWS)
    closesocket(m_socketFd);
    #elif OS(UNIX)
    close(m_socketFd);
    #else
    socketclose(m_socketFd);
    #endif
}

void WebSocketStreamHandlePrivate::socketClosed()
{
    if (m_streamHandle && m_streamHandle->client()) {
        WebSocketStreamHandle* streamHandle = m_streamHandle;
        m_streamHandle = 0;
        // This following call deletes _this_. Nothing should be after it.
        streamHandle->client()->didCloseSocketStream(streamHandle);
    }
}

void WebSocketStreamHandlePrivate::socketError(int error)
{
    if (m_streamHandle && m_streamHandle->client()) {
        WebSocketStreamHandle* streamHandle = m_streamHandle;
        m_streamHandle = 0;
        // This following call deletes _this_. Nothing should be after it.
        streamHandle->client()->didCloseSocketStream(streamHandle);
    }
}

WebSocketStreamHandle::WebSocketStreamHandle(const KURL& url, SocketStreamHandleClient* client)
    : SocketStreamHandleBase(url, client)
{
    bool bError = false;
    int error = WebCore::DomainAccessControl::DOMAIN_ACCESS_OK;
    if (!WebCore::DomainAccessControl::IsDomainAccessible(url.string(), &error))
    {
        switch (error)
        {
        case WebCore::DomainAccessControl::DOMAIN_ACCESS_NO_NETWORK_ACCESS:
			OS::Report("This request cannot be honored. You must enable network access.\n\tRequested URL: %s.\n",
					 encodeWithURLEscapeSequences(url.string()).latin1().data());
			LOG(Network, "This request cannot be honored. You must enable network access.\n\tRequested URL: %s.\n",
				encodeWithURLEscapeSequences(url.string()).latin1().data());
			LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ASSET_LOAD_FAIL, L"", 0);
			break;

		case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ENCRYPTED:
            OS::Report("This request must be over HTTPS. Please use the HTTPS protocol.\n\tRequested URL: %s.\n",
                     encodeWithURLEscapeSequences(url.string()).latin1().data());
            LOG(Network, "This request must be over HTTPS. Please use the HTTPS protocol.\n\tRequested URL: %s.\n",
                encodeWithURLEscapeSequences(url.string()).latin1().data());
            LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_CURLE_UNSUPPORTED_PROTOCOL, L"", 0);
            break;

        case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_WHITELISTED:
            OS::Report("The domain '%s' is not whitelisted. Add it to the Whitelist page of Project Settings.\n",
                     encodeWithURLEscapeSequences(url.string()).latin1().data());
            LOG(Network, "The domain '%s' is not whitelisted. Add it to the Whitelist page of Project Settings.\n",
                encodeWithURLEscapeSequences(url.string()).latin1().data());
            LibPlatform::Error::GetInstance()->Report(LibPlatform::Error::ERROR_CODE_WEBKIT_ADDRESS_BLOCKED_BY_WHITELIST, L"", 0);
            break;

        case WebCore::DomainAccessControl::DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN:
            OS::Report("Before accessing a network resource, an independent service token must first be acquired.\n" \
                    "Please call nwf.act.NintendoAccountManager.acquireNewIndependentServiceToken().\n");
            LOG(Network, "Before accessing a network resource, an independent service token must first be acquired.\n" \
                    "Please call nwf.act.NintendoAccountManager.acquireNewIndependentServiceToken().\n");
            break;

        default:
            OS::Report("This domain is unreachable. Error %d\n", error);
            LOG(Network, "This domain is unreachable. Error %d\n", error);
            break;
        }

        bError = true;
    }

    struct hostent * hostaddr = gethostbyname(url.host().utf8().data());

    if (hostaddr == NULL) {
        OS::Report("could not get IP address for host\n");
        bError = true;
    }
   // int i=0;
   // while ( hostaddr->h_addr_list[i] != NULL) {
   //        OS::Report( "%s ", inet_ntoa( *( struct in_addr*)( hostaddr->h_addr_list[i])));
   //        i++;
   //     }

    int fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

    if (!bError)
    {
        struct sockaddr_in serverAddr;

        memset(&serverAddr, 0, sizeof(serverAddr));
        memcpy(&serverAddr.sin_addr, hostaddr->h_addr_list[0], hostaddr->h_length);

        serverAddr.sin_family = AF_INET;
        if (url.hasPort()) {
            serverAddr.sin_port  = htons(url.port());
        } else {
            serverAddr.sin_port  = htons(80);  //Assume 80 for TCP socket
        }

        if (connect(fd, (struct sockaddr *)&serverAddr, sizeof(struct sockaddr_in)) < 0) {
            bError = true;
            OS::Report("Socket Connect Failed %d\n", errno);
        }
    }

    m_p = new WebSocketStreamHandlePrivate(this, fd);

    m_bConnected = !bError;
}

WebSocketStreamHandle::WebSocketStreamHandle(int fd, SocketStreamHandleClient* client)
    : SocketStreamHandleBase(KURL(), client)
{
    m_p = new WebSocketStreamHandlePrivate(this, fd);
    m_bConnected = (fd >= 0);
}

#if PLATFORM(WIIU) && defined(ENABLE_HIO_INSPECTOR)
WebSocketStreamHandle::WebSocketStreamHandle(int fd, SocketStreamHandleClient* client, bool isHio)
    : SocketStreamHandleBase(KURL(), client)
{
    m_p = new WebSocketStreamHandlePrivate(this, fd, isHio);
}
#endif

WebSocketStreamHandle::~WebSocketStreamHandle()
{
    setClient(0);
    if (m_p)
	    delete m_p;
}

int WebSocketStreamHandle::platformSend(const char* data, int len)
{
    return m_p->socketSend(data, len);
}

void WebSocketStreamHandle::platformClose()
{
    m_p->socketClose();
}



} // namespace WebCore
